"""
TODO: Script to compare CO2SYS.m output to cbsyst output.
Comparison similar to Orr et al (2015)

Parameters
----------
raw_data : str
    Path to GLODAP-WOA comparison data.
co2sys_data : str
    Path to output from CO2SYS.m
output_dir : str
    Output directory for plots

Returns
-------
Prints a range of statistics comparing CO2SYS and cbsyst.
Creates a variety of plots with statistics in output_dir.
"""
